(* set to false for final build *)
val DEBUG=false

structure Geometry =
  struct
    datatype shape =  Rectangle of { x: int, y: int, width: int, height: int }
                                   (* upper left corner *)
                    | Circle of { x: int, y: int, radius: int }
                                (* center *)

    fun dist (x1, y1) (x2, y2) =
        let fun r z = real z
            fun i z = round z
        in
            i(Math.sqrt (Math.pow (r(x2 - x1), 2.0)
                       + Math.pow (r(y2 - y1), 2.0)))
        end

    fun realdist (x1, y1) (x2, y2) =
        Math.sqrt (Math.pow (x2 - x1, 2.0) + Math.pow (y2 - y1, 2.0))

    infix 3 -- ++
    infix 5 ** //

    fun (x1:real, y1:real) -- (x2, y2) = (x1 - x2, y1 - y2)
    fun (x1:real, y1:real) ++ (x2, y2) = (x1 + x2, y1 + y2)
    fun (n:real) ** (x, y) = (n * x, n * y)
    fun (x:real, y:real) // n = (x / n, y / n)

    fun roundvec (x, y) = (round x, round y)

    infix 3 --- +++
    infix 5 *** ///
    fun (x1, y1) --- (x2, y2) = (x1 - x2, y1 - y2)
    fun (x1, y1) +++ (x2, y2) = (x1 + x2, y1 + y2)
    fun (n) *** (x, y) = (n * x, n * y)
    fun (x, y) /// n = (x / n, x / n)

    fun realnorm (x, y) =
        (x, y) // Math.sqrt (x * x + y * y)

    fun inside (Circle {x=cx, y=cy, radius=r}) (x, y)=
            dist (cx, cy) (x, y) < r
      | inside (Rectangle {x=rx, y=ry, width=w, height=h}) (x, y) =
            x >= rx andalso x < rx + w andalso
            y >= ry andalso y < ry + h

    fun intersects (rect1 as Rectangle {x=x1, y=y1, width=w1, height=h1})
                   (rect2 as Rectangle {x=x2, y=y2, width=w2, height=h2}) =
            let (* rectangle 1's corners *)
                val nw1 = (x1, y1)
                val ne1 = (x1 + w1 - 1, y1)
                val sw1 = (x1, y1 + h1 - 1)
                val se1 = (x1 + w1 - 1, y1 + h1 - 1)
                (* rectangle 2's corners *)
                val nw2 = (x2, y2)
                val ne2 = (x2 + w2 - 1, y2)
                val sw2 = (x2, y2 + h2 - 1)
                val se2 = (x2 + w2 - 1, y2 + h2 - 1)
            in
                List.exists (inside rect2) [nw1, ne1, sw1, se1]
                orelse
                List.exists (inside rect1) [nw2, ne2, sw2, se2]
            end
      | intersects (Circle {x=x1, y=y1, radius=r1})
                   (Circle {x=x2, y=y2, radius=r2}) =
            dist (x1, y1) (x2, y2) <= r1 + r2
      | intersects (s1 as Rectangle _) (s2 as Circle _) = intersects s2 s1
      | intersects (circ as Circle {x=cx, y=cy, radius})
                   (rect as Rectangle {x=rx, y=ry, width=w, height=h}) =
            let (* rectangle corners *)
                val nw = (rx, ry)
                val ne = (rx + w - 1, ry)
                val sw = (rx, ry + h - 1)
                val se = (rx + w - 1, ry + h - 1)
                (* circle diamond points, and center *)
                val center = (cx, cy)
                val top = (cx, cy - radius)
                val bot = (cx, cy + radius)
                val left = (cx - radius, cy)
                val right = (cx + radius, cy)
            in
                (* DEBUG *)
                (*
                let fun draw (x, y) = SDL.drawpixel (screen, x, y, red)
                    fun ward (x, y) = SDL.drawpixel (screen, x, y, green)
                in
                    draw center; draw top; draw bot; draw left; draw right;
                    ward nw; ward ne; ward sw; ward se;
                    SDL.flip screen
                end;
                *)
                List.exists (inside circ) [nw, ne, sw, se]
                orelse
                List.exists (inside rect) [top, bot, left, right, center]
            end
  end

open Geometry
infix 3 -- ++
infix 5 ** //

(* SDL.getticks () -- how many hundredths of a second since SDL started? *)
(* event loop: a case statement *)

val screen = SDL.makescreen (700, 700)

fun requireimage s =
      let val s = "testgraphics/" ^ s
      in
          case SDL.Image.load s of
              NONE => (print ("couldn't open " ^ s ^ "\n");
                       raise Fail "end of death")
            | SOME p => p
      end

structure Font =
    FontFn (val surf = requireimage "font.png"
            val charmap =
            " ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789" ^
            "`-=[]\\;',./~!@#$%^&*()_+{}|:\"<>?" (* \" *)
            (* CHECKMARK ESC HEART LCMARK1 LCMARK2 BAR_0 BAR_1 BAR_2 BAR_3-
               BAR_4 BAR_5 BAR_6 BAR_7 BAR_8 BAR_9 BAR_10 BARSTART LRARROW
               LLARROW *)
            val width = 9
            val height = 16
            val styles = 6
            val overlap = 1
            val dims = 3)

(* foreground colors for debugging *)
val red = SDL.color (0wxff, 0w0, 0w0, 0w0)
val green = SDL.color (0w0, 0wxff, 0w0, 0w0)
(* used in message boxes *)
val white = SDL.color (0wxff, 0wxff, 0wxff, 0w0)

(* background color *)
val blue = SDL.color (0w0, 0w0, 0w0, 0w100)

(*
(* XXX DEBUGGING *)
structure SDL =
  struct
    open SDL
    (* fun fillrect _ = () *)
    fun blitall (_, screen, x, y) = fillrect (screen, x, y, 3, 3, red)
  end
*)

fun status s = () (* Font.draw (screen, 10, 10, s) *)

val santamaria = requireimage "santamaria.png"
val monster = requireimage "monster.png"
val island = requireimage "island.png"
val water = requireimage "water1.png"

fun getkey () =
    case SDL.pollevent () of
        NONE => (SDL.delay 0; getkey ())
      | SOME (SDL.E_KeyDown _) => ()
      | SOME _ => (SDL.delay 0; getkey ())

fun messagebox s =
    let val alphagrey = SDL.color (0w100, 0w100, 0w100, 0w50)
        val textwidth = Font.sizex s
        val boxwidth = textwidth + 2 * 50
        val boxheight = 100
        val (x, y) = (SDL.surface_width screen div 2 - boxwidth div 2,
                      SDL.surface_height screen div 2 - boxheight div 2)
        val (tx, ty) = (x + 50, y + 42)
        val (okx, oky) = (x + boxwidth - 62, y + boxheight - 26)
    in
        SDL.fillrect (screen, x+10, y+10, boxwidth, boxheight, alphagrey);
        SDL.fillrect (screen, x, y, boxwidth, boxheight, white);
        Font.draw (screen, tx, ty, s);
        SDL.flip screen;
        SDL.delay 500;
        Font.draw (screen, okx, oky, "[ ok ]");
        SDL.flip screen
    end

fun win () = (messagebox "^0Y^1o^2u ^3m^4a^5d^6e ^7i^8t^9!  ^bAll enemies destroyed!"; getkey ())
fun lose () = (messagebox "^3Oh no!  Your ship is lost!"; getkey ())

fun draw_vector (x, y) (x', y') =
    let in
        SDL.drawpixel (screen, x, y, red);
        SDL.drawpixel (screen, x', y', green)
    end
        handle SDL.SDL _ => ()

fun draw_goal (SOME (Rectangle {x, y, width, height})) =
        let val grey = SDL.color (0w100, 0w100, 0w100, 0w0)
        in
            (* SDL.fillrect (screen, x, y, width, height, grey) *)
            SDL.blitall (island, screen, x, y - 50)
        end
  | draw_goal (_) = ()

datatype bullet = B of { shape: shape, vx: real, vy: real }
datatype enemy = Alive of { shape: shape } (* TODO: add a program *)
               | Destroyed of { shape: shape, ttl: int }

fun Enemy (x, y) = Alive { shape = Circle {x=x, y=y, radius=20} }

fun draw_bullet (B { shape = Circle {x, y, radius}, vx, vy }) =
        SDL.drawcircle (screen, x, y, radius, white)
  | draw_bullet _ = raise Fail "invalid shape for a bullet"

(* XXX slightly misnamed -- initial ttl of destroyed enemies is actually
   initTTL - 1 *)
val initTTL = 30
(* ttlStep is how often (in ticks (XXX portable?)) to update the rendering
   of a destroyed enemy *)
val ttlStep = 3
fun fade 0 surf = surf
  | fade n surf = fade (n-1) (SDL.Util.alphadim surf)
val explosions = Vector.tabulate
                    (initTTL div ttlStep,
                     fn i => fade (initTTL div ttlStep - i) monster)

fun draw_shape (surf, Rectangle {x, y, width=w, height=h}, color) =
            SDL.drawbox (surf, x, y, x+w-1, y+h-1, color)
  | draw_shape (surf, Circle {x, y, radius=r}, color) =
            SDL.drawcircle (surf, x, y, r, color)

fun draw_enemy (Alive {shape = shape as Circle {x, y, radius=rad}}) =
        let val (nwx, nwy) = (x - rad - 10, y - rad - 10)
        in
            SDL.blitall (monster, screen, nwx, nwy);
            if DEBUG then draw_shape (screen, shape, red) else ()
        end
  | draw_enemy (Destroyed { shape = Circle {x, y, radius=rad}, ttl=n }) =
        let val (nwx, nwy) = (x - rad - 10, y - rad - 10)
        in
            (* print ("getting explosion " ^ Int.toString n); *)
            SDL.blitall (Vector.sub (explosions, n div ttlStep), screen, nwx, nwy)
        end

(* XXX DEBUG *)
val forcevector = ref (0.0, 0.0)

datatype state = S of { t : int,            (* clicks since start *)
                        x : real, y : real, (* coords of santamaria *)
                        vx: real, vy: real, (* velocity of santamaria *)
                        ax: real, ay: real, (* accel of santamaria *)
                        lt: bool, rt: bool, (* status of left/right keys *)
                        up: bool, dn: bool, (* status up up/down keys *)
                        spc: bool, spct: int, (* status of spacebar key *)
                        goal: shape option,      (* location of goal *)
                        enemies: enemy list,     (* locations of enemies *)
                        bullets: bullet list     (* locations of bullets *)
                      }

fun draw_background t =
    let val ht = SDL.surface_height
        val wd = SDL.surface_width
        val yr = ref (t mod (ht water) - ht water)
    in
        while !yr < ht screen do
            (let val xr = ref 0
             in
                print ("yr: " ^ Int.toString (!yr) ^ "\n");
                while !xr < wd screen do
                    (SDL.blitall (water, screen, !xr, !yr);
                     xr := !xr + wd water)
             end;
             yr := !yr + ht water)
    end

fun santa_shape (x, y) =
    let val h = SDL.surface_height santamaria - 15
        val w = SDL.surface_width santamaria - 33
    in
        Rectangle {x=x+18, y=y+3, width=w, height=h}
    end

fun draw_santa (x, y) =
     (SDL.blitall (santamaria, screen, x, y);
      if DEBUG then draw_shape (screen, santa_shape (x, y), green) else ())

fun render (state as S {t, x, y, vx, vy, ax, ay,
                        lt, rt, up, dn, spc, spct,
                        goal, enemies, bullets}) =
    let fun r2s r = StringCvt.padLeft #" " 6 (Real.fmt (StringCvt.FIX (SOME 2)) r)
        fun a2s r = StringCvt.padLeft #" " 6 (Real.fmt (StringCvt.FIX (SOME 3)) r)
        fun b2s b = if b then "*" else "_"
    in
    (SDL.clearsurface (screen, blue);
     draw_background t;
     (*
     status ("x=" ^ r2s x ^ ",y=" ^ r2s y
          ^ ",vx=" ^ r2s vx ^ ",vy=" ^ r2s vy
          ^ ",ax=" ^ a2s ax ^ ",ay=" ^ a2s ay
       (* ^ ",^5lt=" ^ b2s lt ^ ",rt=" ^ b2s rt
          ^ ",up=" ^ b2s up ^ ",dn=" ^ b2s dn *)
          ^ " ^4" ^ "fx=" ^ a2s (#1 (!forcevector))
                 ^ ",fy=" ^ a2s (#2 (!forcevector))
          ^ " ^3t=" ^ Int.toString (Word32.toInt t));
     *)
     draw_goal goal;
     app draw_enemy enemies;
     app draw_bullet bullets;
     draw_santa (Real.round x, Real.round y);
     SDL.flip screen)
    end

fun make_state x y goal enemies =
     S { t = 0, x = x, y = y, vx = 0.0, vy = 0.0, ax = 0.0, ay = 0.0,
         lt = false, rt = false, up = false, dn = false, spc = false, spct = 0,
         goal = NONE, enemies = enemies, bullets = [] }

val init_state = make_state 320.0 320.0 NONE []
(*
S { t = 0w0, x = 320.0, y = 320.0, vx = 0.0, vy = 0.0,
                     ax = 0.0, ay = 0.0,
                     lt = false, rt = false, up = false, dn = false,
                     spc = false, spct = 0,
                     goal = NONE, enemies = [], bullets = [] }
*)

val level_one = make_state 100.0 600.0
                (SOME (Rectangle {x=550, y=100, width=100, height=50}))
                [ Enemy (300, 500),
                  Enemy (500, 300),
                  Enemy (475, 475),
                  Enemy (100, 100),
                  Enemy (100, 450),
                  Enemy (400, 100),
                  Enemy (250, 300)(*,
                  Enemy (620, 620)*) ]
    (*S { t = 0w0, x = 100.0, y = 600.0, vx = 0.0, vy = 0.0,
                    ax = 0.0, ay = 0.0,
                    lt = false, rt = false, up = false, dn = false,
                    spc = false, spct = 0,
                    goal = SOME (Rectangle { x=550, y=100, width=100, height=50 }),
                    enemies = [ Enemy (300, 500),
                                Enemy (500, 300),
                                Enemy (475, 475),
                                Enemy (100, 100),
                                Enemy (100, 450),
                                Enemy (400, 100),
                                Enemy (250, 300),
                                Enemy (620, 620) ],
                    bullets = []
                  }
    *)

val level_two = make_state 100.0 600.0
                (SOME (Rectangle { x=550, y=100, width=100, height=50 }))
                [ Enemy (180, 520) ]
    (*S { t = 0w0, x = 100.0, y = 600.0, vx = 0.0, vy = 0.0,
                    ax = 0.0, ay = 0.0,
                    lt = false, rt = false, up = false, dn = false,
                    spc = false, spct = 0,
                    goal = SOME (Rectangle { x=550, y=100, width=100, height=50 }),
                    enemies = [ Enemy (180, 520) ],
                    bullets = []
                  }
    *)

fun setLT (S { t,x,y,vx,vy,ax,ay, lt,rt,up,dn,spc,spct, goal,enemies,bullets }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=V, rt=rt, up=up, dn=dn, spc=spc, spct=spct,
           goal=goal, enemies=enemies, bullets=bullets }
fun setRT (S { t,x,y,vx,vy,ax,ay, lt,rt,up,dn,spc,spct, goal,enemies,bullets }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=lt, rt=V, up=up, dn=dn, spc=spc, spct=spct,
           goal=goal, enemies=enemies, bullets=bullets }
fun setUP (S { t,x,y,vx,vy,ax,ay, lt,rt,up,dn,spc,spct, goal,enemies,bullets }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=lt, rt=rt, up=V, dn=dn, spc=spc, spct=spct,
           goal=goal, enemies=enemies, bullets=bullets }
fun setDN (S { t,x,y,vx,vy,ax,ay, lt,rt,up,dn,spc,spct, goal,enemies,bullets }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=lt, rt=rt, up=up, dn=V, spc=spc, spct=spct,
           goal=goal, enemies=enemies, bullets=bullets }
fun setSP (S { t,x,y,vx,vy,ax,ay, lt,rt,up,dn,spc,spct, goal,enemies,bullets }) V =
       S { t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
           lt=lt, rt=rt, up=up, dn=dn, spc=V, spct=spct,
           goal=goal, enemies=enemies, bullets=bullets }

(*
fun whirlpool_force (x, y) (Circle {x=cx, y=cy, ...}) =
    let val radius = SDL.surface_height santamaria div 2
        val (sx, sy) = (x + real radius, y + real radius)
        val (cx, cy) = (real cx, real cy)
        val r = realdist (cx, cy) (sx, sy)
        (* val magin = 0.45 / (r * r) (* inverse square law *) *)
        val dirin = realnorm ((cx, cy) -- (sx, sy))
        (* val magtan = 0.1 * magin (* 0.000001 * Math.ln r (* if r < 40.0 then 0.005 / r else 0.0 (* 0.5 * magin *) *) *) *)
        (* for velocity instead of acceleration -- can't get the constants
           to be quite right :/ *)
        val magin = if r < 100.0 then 0.75 / r else 0.0
        val magtan = 0.3 * magin * Math.pow (r, 0.2)
        val dirtan = (~(#2 dirin), #1 dirin)
    in
        (magin ** dirin) ++ (magtan ** dirtan)
    end
*)

fun update_bullet (B { shape = Circle {x, y, radius}, vx, vy}) =
    let (* fun r x = Real.round x *)
        fun inbounds_x x = x >= 0 andalso x < SDL.surface_width screen
        fun inbounds_y y = y >= 0 andalso y < SDL.surface_height screen

        val newx = Real.round (real(x) + vx)
        val newy = Real.round (real(y) + vy)
    in
        if inbounds_x newx andalso inbounds_y newy then
            SOME (B { shape = Circle {x=newx, y=newy, radius=radius},
                      vx = vx, vy = vy })
        else
            NONE
    end

fun new_bullet (x, y) =
    B { shape=Circle {x=Real.round x, y=Real.round y, radius=3},
        vx = 0.0, vy = ~6.0 }

(*
fun collides (Circle {x=x1, y=y1, radius=r1}) (Circle {x=x2, y=y2, radius=r2}) =
    dist (x1, y1) (x2, y2) <= r1 + r2

fun collides_rect (Circle {x=cx, y=cy, radius=cradius})
                  (Rectangle {x=rx, y=ry, width=rwidth, height=rheight})
*)

fun collide_enemy bullets (enemy as Alive { shape=ecirc, ... }) =
        if List.exists (intersects ecirc) (map (fn B {shape=c,...} => c) bullets)
        then SOME (Destroyed { shape = ecirc, ttl = initTTL-1 })
        else SOME enemy
  | collide_enemy _ (Destroyed { shape=_, ttl=0 }) = NONE
  | collide_enemy _ (Destroyed { shape=c, ttl=n }) = SOME (Destroyed {shape=c,
                                                                      ttl=n-1})
fun enemy_shape (Alive e) = SOME (#shape e)
  | enemy_shape _ = NONE

(* ditto *)
fun collide_bullet enemies (B bullet) =
    if List.exists (intersects (#shape bullet))
                   (List.mapPartial enemy_shape enemies)
    then NONE
    else SOME (B bullet)

fun update (state as S {t, x, y, vx, vy, ax, ay,
                        lt, rt, up, dn, spc, spct,
                        goal, enemies, bullets}) =
    let fun r x = Real.round x
        fun inbounds_x x = r x >= 0 andalso r x < SDL.surface_width screen
        fun inbounds_y y = r y >= 0 andalso r y < SDL.surface_height screen

        (*** move boat ***)

        (* XXX 3.0 is a magic constant -- is this portable? *)
        val vx = if lt andalso not rt then ~3.0
                else if rt andalso not lt then 3.0
                else 0.0
        val vy = if up andalso not dn then ~3.0
                else if dn andalso not up then 3.0
                else 0.0
        (*
        (* XXX DEBUG *)
        val () = forcevector := (fx, fy)
        *)
        val newx = x + vx
        val newy = y + vy
        val (x, vx) = if inbounds_x newx then (newx, vx) else (x, vx)
        val (y, vy) = if inbounds_y newy then (newy, vy) else (y, vy)

        (*** move bullets ***)

        val bullets = List.mapPartial update_bullet bullets

        (*** create bullets ***)

        val bullets = if spc andalso spct = 0
                      then let val bx = x + real(SDL.surface_width santamaria div 2)
                           in new_bullet (bx, y)::bullets
                           end
                      else bullets
        (* spct is the spacebar cooldown timer -- keeps held down spacebar
           from firing too fast.  XXX 16 is slightly magic -- portable? *)
        fun cool x = if x <= 0 then 16 else x - 1
        val spct = if not spc then 0 else cool spct

        (*** destroy enemies ***)

        val newenemies = List.mapPartial (collide_enemy bullets) enemies
        val newbullets = List.mapPartial (collide_bullet enemies) bullets

        (*** update timer ***)
        val t = t + 1
    in
      (* XXX time is not doing anything at the moment *)
      S {t=t, x=x, y=y, vx=vx, vy=vy, ax=ax, ay=ay,
         lt=lt, rt=rt, up=up, dn=dn, spc=spc, spct=spct,
         goal=goal, enemies=newenemies, bullets=newbullets}
    end

fun winning (state as S {x, y, goal = _ (* SOME (goal as Rectangle _) *),
                         t, vx, vy, ax, ay,
                         lt, rt, up, dn, spc, spct,
                         enemies, bullets }) =
        let val h = SDL.surface_height santamaria
            val w = SDL.surface_width santamaria
            val santa = santa_shape (round x, round y)
        in
            enemies = [] (* andalso
            intersects santa goal *)
        end
  | winning _ = false

fun losing (state as S {x, y, enemies, ...}) =
    List.exists (intersects (santa_shape (round x, round y)))
                (List.mapPartial enemy_shape enemies)

fun loop state =
    let
        (* val state as S {t, x, y, vx, vy, ax, ay, lt, rt, up, dn, goal, enemies} = update state *)
        val state = update state
        val () = render state
    in
    if winning state then
        (win (); loop level_one)
    else if losing state then
        (lose (); loop level_one)
    else
    case SDL.pollevent () of
        NONE => (SDL.delay 0; loop state)
      (* quit on escape or Mac quit *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_ESCAPE }) => ()
      | SOME SDL.E_Quit => ()
      (* win on w! *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_w }) => win ()
      (* change dx and dy on arrow keys *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_LEFT })  => loop (setLT state true)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_RIGHT }) => loop (setRT state true)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_UP })    => loop (setUP state true)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_DOWN })  => loop (setDN state true)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_SPACE }) => loop (setSP state true)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_LEFT })  => loop (setLT state false)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_RIGHT }) => loop (setRT state false)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_UP })    => loop (setUP state false)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_DOWN })  => loop (setDN state false)
      | SOME (SDL.E_KeyUp { sym = SDL.SDLK_SPACE }) => loop (setSP state false)
      (* levels *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_0 }) => loop init_state
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_1 }) => loop level_one
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_2 }) => loop level_two
      (* pause *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_p }) => (getkey (); loop state)
      | SOME _ => loop state
    end

val () = (messagebox "^51492:^3 Destroy the sea monsters to discover the new world!";
          getkey ();
          (* loop init_state *)
          loop level_one)
